# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Creates initializer function from flags."""
from torch import nn


def initialize(flag_init: str) -> nn.init:
  """Defines the initializer for the model parameters."""
  if flag_init == 'normal':
    initializer = nn.init.normal_
  elif flag_init == 'uniform':
    initializer = nn.init.uniform_
  elif flag_init == 'xavier_normal':
    initializer = nn.init.xavier_normal_
  elif flag_init == 'xavier_uniform':
    initializer = nn.init.xavier_uniform_
  elif not flag_init:
    initializer = nn.init.uniform_
  else:
    raise ValueError('Initializer not provided.')
  return initializer

